#ifndef SHARED_PTR
#define SHARED_PTR

#include <map>
#include <iostream>

struct A {
	static int counter;
	int i;
	A() : i(counter++) { std::cout << "A constructor" << std::endl; }
	~A() { std::cout << "A destructor" << std::endl; }
};

int A::counter=0;

template<class T>
std::map<T*, int>& getMap()
{
	static std::map<T*, int> ref_counters;
	return ref_counters;
}

// Now, reference counting
template<class T>
class SharedPointer {
	static std::map<T*, int> ref_counters;
	T *ptr;
public:
	SharedPointer(T *p) : ptr(p)
	{
		// see if in map
		typename std::map<T*, int>::iterator elem = ref_counters.find(ptr);
		if (elem == ref_counters.end()) ref_counters[ptr] = 1;
		else elem->second++;
	}
	~SharedPointer() {
		typename std::map<T*, int>::iterator elem = ref_counters.find(ptr);
		if (elem == ref_counters.end()) // ?? Why cannot be found ??
			throw 101;
		if ((--ref_counters[ptr]) == 0) delete ptr;
	}

	SharedPointer(const SharedPointer &p) {
		ptr = p.ptr;
		if ( ptr != 0 ) ref_counters[ptr]++;
	}

	SharedPointer &operator=(const SharedPointer &p) {
		if (ptr != p.ptr) {
			if (ptr != 0) {
				if (--ref_counters[ptr] == 0) delete ptr;
			}
			ptr = p.ptr;
			ref_counters[ptr]++;
		}
		return *this;
	}

	T * operator->() {
		return ptr;
	}

	T & operator*() {
		return *ptr;
	}
};

template<class T>
std::map<T*, int> SharedPointer<T>::ref_counters;

#endif
